@@ -8,14 +8,16 @@ from pathlib import Path
 import sys
 from typing import Union, Dict, List
 
+from pypaq.exception import PyPaqException
+
 
-# ********************************************* for raise_exception=False each reader will return None if file not found
 
 # pickle read
 def r_pickle(
         file_path,
-        obj_type=           None, # if obj_type is given checks for compatibility with given type
-        raise_exception=    False):
+        obj_type=           None,   # if obj_type is given checks for compatibility with given type
+        raise_exception=    False,  # for False returns None if file not found
+):
     if not os.path.isfile(file_path):
         if raise_exception: raise FileNotFoundError(f'file {file_path} not exists!')
         return None
@@ -23,7 +25,9 @@ def r_pickle(
     # obj = pickle.load(open(file_path, 'rb')) << replaced by:
     with open(file_path, 'rb') as file: obj = pickle.load(file)
 
-    if obj_type: assert type(obj) is obj_type, f'ERROR: obj from file is not {str(obj_type)} type !!!'
+    if obj_type:
+        if not type(obj) is obj_type:
+            raise PyPaqException(f'ERROR: obj from file is not {str(obj_type)} type')
     return obj
 
 # pickle write
@@ -36,7 +40,8 @@ def w_pickle(
 # json read
 def r_json(
         file_path,
-        raise_exception=    False):
+        raise_exception=    False,  # for False returns None if file not found
+):
     if not os.path.isfile(file_path):
         if raise_exception: raise FileNotFoundError(f'file {file_path} not exists!')
         return None
@@ -53,7 +58,8 @@ def w_json(
 # jsonl read
 def r_jsonl(
         file_path,
-        raise_exception=False):
+        raise_exception=    False,  # for False returns None if file not found
+):
     if not os.path.isfile(file_path):
         if raise_exception: raise FileNotFoundError(f'file {file_path} not exists!')
         return None
@@ -72,7 +78,8 @@ def w_jsonl(
 # csv read
 def r_csv(
         file_path,
-        raise_exception=    False):
+        raise_exception=    False,  # for False returns None if file not found
+):
     if not os.path.isfile(file_path):
         if raise_exception: raise FileNotFoundError(f'file {file_path} not exists!')
         return None
