@@ -1,12 +1,15 @@
 from typing import Optional
 
+from pypaq.exception import PyPaqException
+
+
 # moving average class, updates self.value with factor (given while init)
 class MovAvg:
 
     def __init__(
             self,
-            factor=     0.1,
-            first_avg=  True,   # first 1/factor/2 values will be averaged
+            factor: float=      0.1,    # (0.0;1.0>
+            first_avg: bool=    True,   # first (1/factor/2) values will be averaged
     ):
         self.value: Optional[float] = None
         self.factor = factor
@@ -14,6 +17,7 @@ class MovAvg:
         self.first_avg = first_avg
         self.firstL = []
 
+
     def upd(self, val:float):
 
         if self.first_avg and self.upd_ix < 1/self.factor/2:
@@ -27,10 +31,14 @@ class MovAvg:
 
         return self.value
 
+
     def reset(self, val:Optional[float]=None):
         self.value = val
         self.upd_ix = 0
         self.firstL = []
 
+
     def __call__(self) -> float:
+        if self.value is None:
+            raise PyPaqException('MovAvg not updated yet, value unknown')
         return self.value
\ No newline at end of file
