@@ -1,24 +1,19 @@
-"""
-
- 2019 (c) piteren
-
-    stats methods
-
-"""
-
 import pandas as pd
 import numpy as np
+from typing import Dict, Tuple
+
+from pypaq.pytypes import NPL
 
 
 # min, avg, max ..of num list
-def mam(vals: list):
-    if vals: return [min(vals), sum(vals) / len(vals), max(vals)]
-    else:    return [0,0,0]
+def mam(vals:NPL) -> Tuple[float,float,float]:
+    if len(vals): return min(vals), sum(vals) / len(vals), max(vals)
+    else:         return 0.0, 0.0, 0.0
 
 # mean, std, min, max (from given list of values or np.arr)
-def msmx(vals : list or np.array) -> dict:
+def msmx(vals:NPL) -> Dict:
 
-    arr = np.array(vals) if type(vals) is list else vals
+    arr = np.asarray(vals) if type(vals) is list else vals
     ret_dict = {
         'mean': float(np.mean(arr)),
         'std':  float(np.std(arr)),
@@ -29,7 +24,7 @@ def msmx(vals : list or np.array) -> dict:
 
 # deep stats (with pandas)
 def stats_pd(
-        val_list :list,
+        val_list :NPL,
         n_percentiles=  10) -> str:
     s = f'{pd.Series(val_list).describe(percentiles=[0.1*n for n in range(1,n_percentiles)])}'
     return s[:s.rfind('\n')]
\ No newline at end of file
