@@ -5,15 +5,21 @@ import psutil
 import time
 from typing import Any, Optional
 
+from pypaq.exception import PyPaqException
 from pypaq.lipytools.pylogger import get_pylogger
 
 
+class MPythonException(PyPaqException):
+    pass
+
+
 # message sent between processes via Ques (my que)
 class QMessage:
     def __init__(self, type:str, data:Any):
         self.type = type
         self.data = data
 
+
 # https://github.com/vterron/lemon/commit/9ca6b4b1212228dbd4f69b88aaf88b12952d7d6f
 class SharedCounter:
 
@@ -28,6 +34,7 @@ class SharedCounter:
     def value(self) -> int:
         return self.count.value
 
+
 # my que
 class Que:
 
@@ -36,7 +43,8 @@ class Que:
         self.size = SharedCounter(0)
 
     def put(self, msg:QMessage, **kwargs):
-        assert isinstance(msg, QMessage)
+        if not isinstance(msg, QMessage):
+            raise MPythonException(f'\'msg\' should be type of QMessage, but is {type(msg)}')
         self.size.increment(1)
         self.q.put(msg, **kwargs)
 
@@ -48,7 +56,8 @@ class Que:
         try:
             msg = self.q.get(block=block, timeout=timeout)
             self.size.increment(-1)
-            assert isinstance(msg, QMessage)
+            if not isinstance(msg, QMessage):
+                raise MPythonException(f'\'msg\' should be type of QMessage, but is {type(msg)}')
             return msg
         except Empty:
             return None
@@ -70,7 +79,7 @@ class ExSubprocess(Process, ABC):
             id: Optional[int or str]=   None,   # unique id to identify the subprocess, if not given takes from Process name
             raise_unk_exception=        True,   # raises exception other than KeyboardInterrupt
             logger=                     None,
-            loglevel=                   20):
+            loglevel=                   30):
 
         ABC.__init__(self)
         Process.__init__(self, target=self.__run)
@@ -101,7 +110,7 @@ class ExSubprocess(Process, ABC):
         except KeyboardInterrupt:
             self.__exception_handle('KeyboardInterrupt')
         except Exception as e:
-            self.__exception_handle(f'other: {str(e)}')
+            self.__exception_handle(f'other: {e}')
             if self.raise_unk_exception: raise e
 
     # method run in subprocess, to be implemented
@@ -137,10 +146,13 @@ class ExSubprocess(Process, ABC):
     def closed(self):
         return self._closed
 
+    # some process info
     def get_info(self) -> str:
-        pid = self.pid if not self.closed else '<pid:closed>'
+        pid = self.pid if not self.closed else None
+        pid_nfo = pid if pid is not None else '<pid:closed>'
         exitcode = self.exitcode if not self.closed else '<exitcode:closed>'
-        nfo = f'{str(self)}, pid: {pid}, parent pid: {self._parent_pid}, alive: {self.alive}, exitcode: {exitcode}'
+        mem_nfo = int(psutil.Process(pid).memory_info().rss / 1024 ** 2) if not exitcode else '-'
+        nfo = f'{str(self)}, pid: {pid_nfo}, mem:{mem_nfo}MB, parent pid: {self._parent_pid}, alive: {self.alive}, exitcode: {exitcode}'
         return nfo
 
 
