@@ -2,14 +2,15 @@ import os
 import shutil
 from typing import Optional, List
 
+from pypaq.exception import PyPaqException
 from pypaq.lipytools.printout import stamp
 from pypaq.lipytools.files import r_pickle, w_pickle, prep_folder
-from pypaq.lipytools.pylogger import get_pylogger, get_child
+from pypaq.lipytools.pylogger import get_pylogger
 from pypaq.pms.base import POINT
 from pypaq.pms.subscriptable import SubGX
 
 
-class ParaSaveException(Exception):
+class ParaSaveException(PyPaqException):
     pass
 
 
@@ -22,8 +23,7 @@ class ParaSave(SubGX):
 
     PARASAVE_DEFAULTS = {
         'gxable':   True,
-        'parents':  [], # list of parents names TODO: what is this for?
-    }
+        'parents':  []} # list of parents names
 
     SAVE_TOPDIR = None
     SAVE_FN_PFX = 'point'   # POINT file prefix
@@ -57,7 +57,6 @@ class ParaSave(SubGX):
         self.__log.debug(f'> save_topdir: {self.save_topdir}')
         self.__log.debug(f'> save_fn_pfx: {self.save_fn_pfx}')
 
-
         if assert_saved and not os.path.isfile(ParaSave.__obj_fn(name, self.save_topdir, self.save_fn_pfx)):
             ex_msg = f'ParaSave {self.name} does not exist!'
             self.__log.error(ex_msg)
